/*
 * z2-Environment
 * 
 * Copyright(c) ZFabrik Software GmbH & Co. KG
 * 
 * contact@zfabrik.de
 * 
 * http://www.z2-environment.eu
 */
package com.zfabrik.impl.db.data;

import java.lang.reflect.InvocationHandler;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Proxy;
import java.sql.Connection;
import java.sql.SQLException;
import java.sql.SQLFeatureNotSupportedException;
import java.util.Properties;
import java.util.concurrent.Callable;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.naming.InitialContext;
import javax.sql.DataSource;

import com.mchange.v2.c3p0.ComboPooledDataSource;
import com.mchange.v2.c3p0.DataSources;
import com.mchange.v2.c3p0.impl.IdentityTokenResolvable;
import com.zfabrik.components.IComponentDescriptor;
import com.zfabrik.components.IComponentsLookup;
import com.zfabrik.components.IComponentsManager;
import com.zfabrik.components.java.IJavaComponent;
import com.zfabrik.components.java.JavaComponentUtil;
import com.zfabrik.resources.IResourceHandle;
import com.zfabrik.resources.ResourceBusyException;
import com.zfabrik.resources.provider.Resource;
import com.zfabrik.util.PropertyExpressionResolver;
import com.zfabrik.util.threading.ThreadUtil;
import com.zfabrik.work.IWorkResource;
import com.zfabrik.work.WorkException;
import com.zfabrik.work.WorkUnit;

/**
 * Implementation of component type <b>javax.sql.DataSource</b>. This implementation supports a pooled data source implementation that 
 * integrates with Z2's unit of work mechanism (see {@link WorkUnit}).
 * <p>
 * A component of type <b>javax.sql.DataSource</b> can be configured in various ways, controlled by the following 
 * parameters:
 * 
 * <dl>
 * <dt>ds.type</dt><dd>This can be either <b>ZFabrikPoolingDataSource</b> or <b>NativeDataSource</b>. In the first case, the data source will be pooled 
 * using {@link PoolingDataSource}. In that case the pool configuration parameters in {@link PoolingDataSource} should be applied</dd>
 * 
 * <dt>ds.dataSourceClass</dt><dd>If set, the specified data source class will be loaded as data source implementation using the private class loader
 * of the Java module of the component holding the data source definition. When specifying this class in conjunction with the <code>ds.type=ZFabrikPoolingDataSource</code>
 * configuration properties will be applied to both and the pool will request new connections from the specified data source. Alternatively, the pool may be configured
 * to use a driver class. See {@link PoolingDataSource} for more details.</dd>
 * 
 * <dt>ds.dataSourceJNDIName</dt><dd>If <code>ds.dataSourceClass</code> is not set, this JNDI name will be tried to lookup a data source that will
 * serve as connection provider. During lookup the thread context class loader will be set to the private class loader
 * of the Java module of the component holding the data source definition. When specifying this class in conjunction with the <code>ds.type=ZFabrikPoolingDataSource</code>
 * configuration properties will be applied to both and the pool will request new connections from the specified data source. Alternatively, the pool may be configured
 * to use a driver class. See {@link PoolingDataSource} for more details.</dd>
 * 
 * <dt>ds.enlist</dt><dd>This can be either <b>workUnit</b> or <b>none</b>. If set to <b>workUnit</b>, connections will be enlisted with the 
 * {@link WorkUnit} and only committed, rolled back, or closed when the unit of work was completed. Transparently this integrates with Z2's 
 * pseudo-distributed JTA implementation. Please check the <a href="http://redmine.z2-environment.net">Wiki</a> at  for more information.
 * If set to <b>none</b> connections will not be enlisted but simply handed out from the data source.</dd>
 * <dt>ds.prop.abc</dt><dd>Attempts to set the bean property <b>abc</b> on the data source (and on PoolingDataSource, if used). A setter method <code>setAbc</code>
 * will be used for that. The property <b>ds.propType.abc</b> defines the signature required.</dd>
 * <dt>ds.propType.abc</dt><dd>Defines the type of a configuration property as in <b>ds.prop.abc</b>. Can be any of <b>int</b>,<b>string</b>, or <b>boolean</b>. Defaults to <b>string</b>.</dd>
 * 
 * @author hb
 */
public class DataSourceResource extends Resource {
	
	public final static boolean IS_JAVA7 = Float.parseFloat(System.getProperty("java.specification.version"))>1.6f;	
	
	private static final Class<?> INT_SIGNATURE = int.class;
	private static final Class<?> STRING_SIGNATURE = String.class;
	private static final Class<?> BOOLEAN_SIGNATURE = boolean.class;

		
	private static final String STRING = "string";
	private static final String INT = "int";
	private static final String BOOLEAN = "boolean";
	private static final String DS_PROP_TYPE_PREFIX = "ds.propType.";
	private static final String DS_PROP_PREFIX = "ds.prop.";

	private static final String DS_TEMPLATE_COMPONENT = "ds.templateComponent";
	private static final String DS_TYPE = "ds.type";
	private static final String DS_CLASS = "ds.dataSourceClass";
	private static final String DS_JNDI = "ds.dataSourceJNDIName";

	private static final String TYPE_COMBO   = "ComboPooledDataSource";
	private static final String TYPE_NATIVE  = "NativeDataSource";
	private static final String TYPE_ZFABRIK = "ZFabrikPoolingDataSource";

	private static final String DS_ENLIST = "ds.enlist";
	private static final String DS_ENLIST_WORKUNIT = "workUnit"; // enlist in
	// workunit.
	private static final String DS_ENLIST_NONE = "none"; // not enlisting.

	// Java 6/7 compatibility helper
	static Method getMethodIfAvailable(Class<?> clz, String methodName, Class<?> ... argsClz) {
		Method m = null;
		if (IS_JAVA7) {
			try {
				m = clz.getMethod(methodName,argsClz);
			} catch (NoSuchMethodException nsme) {
				if (logger.isLoggable(Level.FINE)) {
					logger.fine("Method "+methodName + " not found on "+clz+" (which may not be prepared for Java 7)");
				}
			} catch (Exception e) {
				throw new RuntimeException("Failed to retrieve Java 7 method "+methodName+" on "+clz,e);
			}
		}
		return m;
	}
	
	
	// helper method for Java 6/7 compatibility on DataSource
	static Logger getParentLogger(Object instance, Method method) throws SQLFeatureNotSupportedException {
		if (method!=null) {
			try {
				return (Logger) method.invoke(instance);
			} catch (InvocationTargetException e) {
				if (e.getCause() instanceof SQLFeatureNotSupportedException) {
					throw (SQLFeatureNotSupportedException) e.getCause();
				}
				throw new RuntimeException(e);
			} catch (Exception e) {
				throw new RuntimeException(e);
			}
		}
    	throw new UnsupportedOperationException();
	}

	
	/**
	 * Work resource, keeping the connection per data source on the work unit.
	 * @author hb
	 *
	 */
	class DSWU implements IWorkResource {
		private Connection conn;
		private Connection wrapper;

		public DSWU(final Connection conn) throws Exception {
			this.conn = conn;
			
			// overwrite the close method. 
			this.wrapper = (Connection) Proxy.newProxyInstance(this.getClass().getClassLoader(), new Class<?>[]{Connection.class}, new InvocationHandler() {
				public Object invoke(Object proxy, Method method, Object[] args) throws Throwable {
					if (!"close".equals(method.getName())) {
						return method.invoke(DSWU.this.conn, args);
					}
					return null;
				}
			});
		}

		public Connection getConnection() {
			return this.wrapper;
		}

		public void begin() {
			if (logger.isLoggable(Level.FINER)) {
				logger.finer("Begin connection \""+this.wrapper+"\" : "+name);
			}
			try {
				this.conn.setAutoCommit(false);
			} catch (SQLException e) {
				throw new WorkException("Work unit begin failed: "+name);
			}
		}
		public void beforeCompletion(boolean rollback) {}
		public void afterCompletion(boolean rollback) {}
		public void close() throws WorkException {
			try {
				if (logger.isLoggable(Level.FINER)) {
					logger.finer("Close connection \""+this.wrapper+"\": "+name);
				}
				this.conn.close();
			} catch (SQLException e) {
				throw new WorkException("Error during close of connection: " + name, e);
			}
			this.conn = null;
		}

		public void commit() {
			if (logger.isLoggable(Level.FINER)) {
				logger.finer("Commit connection \""+this.wrapper+"\" : "+name);
			}
			if (this.conn != null) {
				try {
					if (!this.conn.isClosed()) {
						this.conn.commit();
					}
				} catch (SQLException e) {
					throw new WorkException("Error during commit: " + name, e);
				}
			}
		}

		public void rollback() throws WorkException {
			if (logger.isLoggable(Level.FINER)) {
				logger.finer("Rollback connection \""+this.wrapper+"\" : "+name);
			}
			if (this.conn != null) {
				try {
					if (!this.conn.isClosed()) {
						this.conn.rollback();
					}
				} catch (SQLException e) {
					throw new WorkException("Error during rollback: " + name, e);
				}
			} 
		}

	}
	
	//

	String name;
	private DataSourceWrapper ds;
	boolean enlistInWorkUnit;
	private Properties origProps;
	private Properties resolvedProps;
	// house keeping!
	private IResourceHandle tplh;

	public DataSourceResource(String name) {
		this.name = name;
	}

	public synchronized <T> T as(Class<T> clz) {
		if (DataSource.class.equals(clz) || Object.class.equals(clz)) {
			_load();
			return clz.cast(this.ds);
		} else 
		if (Properties.class.equals(clz)) {
			_loadProps();
			// return the unresolved properties! So that referring data sources may provide their own overrides
			return clz.cast(this.origProps);
		}
		return null;
	}

	private  void _loadProps() {
		if (this.ds == null) {
			IComponentDescriptor desc = IComponentsManager.INSTANCE.getComponent(this.name);
			if (desc == null)
				throw new IllegalStateException("Component descriptor not found: " + this.name);

			Properties props = new Properties();

			// check for a template.
			String tplc = desc.getProperties().getProperty(DS_TEMPLATE_COMPONENT);
			if (tplc!=null) {
				// yepp... got one. Make ourselfs dependent and mix the props!
				this.tplh = IComponentsLookup.INSTANCE.lookup(tplc,IResourceHandle.class);
				Properties tplp = this.tplh.as(Properties.class);
				if (tplp==null) {
					throw new IllegalStateException("Template component ("+tplc+") for data source not found: "+this.name);
				}
				handle().addDependency(tplh);
				// start with the template
				props.putAll(tplp);
			}
			// and with higher pref. the local props
			props.putAll(desc.getProperties());
			this.origProps = props;
			this.resolvedProps = new Properties();
			// resolve all expressions
			new PropertyExpressionResolver(this.origProps,this.resolvedProps).resolveAll();
		}
	}
	
	
	protected static ClassLoader getClassLoader(String name) {
    	IJavaComponent jc = JavaComponentUtil.getJavaComponent(name).as(IJavaComponent.class);
    	if (jc==null) {
    		return DataSourceResource.class.getClassLoader();
    	} else {
    		return jc.getPrivateLoader();
    	}
	}
	
	private  void _load() {
		if (this.ds == null) {
			this._loadProps();
		
			String ts = this.resolvedProps.getProperty(DS_TYPE);
			if (ts == null) {
				throw new IllegalStateException("Must specify data source type (" + DS_TYPE + " in {" + TYPE_COMBO + ", " + TYPE_NATIVE + ", "+TYPE_ZFABRIK+"}): "
						+ this.name);
			}
			final String type = ts.trim();

			String enl = this.resolvedProps.getProperty(DS_ENLIST);
			if (enl != null) {
				enl = enl.trim();
				if (DS_ENLIST_WORKUNIT.equals(enl)) {
					this.enlistInWorkUnit = true;
				} else if (DS_ENLIST_NONE.equals(enl)) {
					this.enlistInWorkUnit = false;
				} else
					throw new IllegalStateException("Property " + DS_ENLIST + " must be in {" + DS_ENLIST_NONE + "*, " + DS_ENLIST_WORKUNIT + "}: "
							+ this.name);
			}

			DataSource lds = null;
			try {
				lds = ThreadUtil.cleanContextExceptionExecute(new Callable<DataSource>() {
					public DataSource call() throws Exception {
						// data source to keep and data source to configure as well
						DataSource ds,dc;
						if (TYPE_COMBO.equals(type)) {
							ComboPooledDataSource cds = new ComboPooledDataSource(false);
							cds.setIdentityToken(name);
							IdentityTokenResolvable.doResolve(cds);
							dc=ds=cds;
						} else if (TYPE_NATIVE.equals(type)) {
							DataSource dds = createDataSource();
							if (dds == null) {
								throw new IllegalStateException("Must specify data source (" + DS_CLASS + " or "+ DS_JNDI+") when using type " + TYPE_NATIVE	+ ": " + name);
							}
							ds=dc=dds;
						} else if (TYPE_ZFABRIK.equals(type)) {
							DataSource dds = createDataSource();
							if (dds!=null) {
								dc= dds;
								ds= new PoolingDataSource(handle(), name, dc);
							} else {
								dc=ds= new PoolingDataSource(handle(), name);
							}
							
						} else {
							throw new IllegalStateException("Unsupported data source type (" + DS_TYPE + "=" + type + "): " + name);
						}
						for (Object k : resolvedProps.keySet()) {
							String na = (String) k;
							if (na.startsWith(DS_PROP_PREFIX)) {
								na = na.substring(DS_PROP_PREFIX.length());
								String t = resolvedProps.getProperty(DS_PROP_TYPE_PREFIX + na, STRING);
								if (t != null) {
									t = t.toLowerCase().trim();
								}
								String methodName = "set" + Character.toUpperCase(na.charAt(0)) + na.substring(1);
								String val = resolvedProps.getProperty((String) k);
								if (val != null) {
									val = val.trim();
								} else {
									throw new IllegalArgumentException("Datasource property " + na + " has no value: " + name);
								}
								if (INT.equals(t)) {
									invokeSetter(methodName, INT_SIGNATURE, Integer.parseInt(val), dc,ds);
								} else if (STRING.equals(t)) {
									invokeSetter(methodName, STRING_SIGNATURE, val, dc,ds);
								} else if (BOOLEAN.equals(t)) {
									boolean v = Boolean.parseBoolean(val);
									invokeSetter(methodName, BOOLEAN_SIGNATURE, v, dc,ds);
								} else {
									throw new IllegalArgumentException("unsupported property type " + t);
								}
							}
						}
						return ds;
					}

					// create a data source
					// by class name or by jndi name
					private DataSource createDataSource() {
						return ThreadUtil.cleanContextExecute(
							getClassLoader(name),
							new Callable<DataSource>() {
								@Override
								public DataSource call() throws Exception {
									// try by class name first
									String csclz = resolvedProps.getProperty(DS_CLASS,"").trim();
									if (csclz.length()>0) {
										try {
											return (DataSource) Class.forName(csclz.trim(),false,getClassLoader(name)).newInstance();
										} catch (Exception e) {
											throw new IllegalStateException("Failed to instantiate data source class "+csclz,e);
										}
									}
									// try by component name next
									String jndi = resolvedProps.getProperty(DS_JNDI, "").trim();
									if (jndi.length()>0) {
										return (DataSource) InitialContext.doLookup(jndi);
									}
									return null;
								}
							}
						);
					}

					// sets on all objects that provide the setter
					private void invokeSetter(String methodName, Class<?> signature, Object value, Object ... targets) 	throws Exception {
						for (Object o : targets) {
							try {
								o.getClass().getMethod(methodName, signature).invoke(o, value);
							} catch (NoSuchMethodException nsme) {
								// that's ok. We try the next.
							}
						}
					}
				});

				// wrap
				this.ds = new DataSourceWrapper(this, lds);
			} catch (Exception e) {
				throw new IllegalStateException("Failed to configure data source: " + this.name, e);
			} finally {
				if (this.ds == null) {
					// something failed
					_destroy(lds);
				}
			}
		}
	}

	void _destroy(DataSource ds) {
		if (ds instanceof PoolingDataSource) {
			((PoolingDataSource)ds).destroy();
		} else
		if (ds instanceof ComboPooledDataSource) {
			try {
				((ComboPooledDataSource) ds).close();
			} catch (Exception e) {
				logger.log(Level.WARNING,"Caught exception when closing pooled data source: "+name,e);
			} finally {
				try {
					DataSources.destroy((ComboPooledDataSource) ds);
				} catch (Exception e) {
					logger.log(Level.WARNING,"Caught exception when destroying pooled data source: "+name,e);
				}
			}				
		}
	}
	
	public synchronized void invalidate() throws ResourceBusyException {
		if (this.ds != null) {
			try {
				this.ds._kill();
			} finally {
				this.ds = null;
				this.resolvedProps=null;
				this.origProps=null;
				this.tplh = null;
			}
		}
	}

	final static Logger logger = Logger.getLogger(DataSourceResource.class.getName());
}
