/*
 * z2-Environment
 * 
 * Copyright(c) ZFabrik Software GmbH & Co. KG
 * 
 * contact@zfabrik.de
 * 
 * http://www.z2-environment.eu
 */
package com.zfabrik.tx;

import java.util.concurrent.Callable;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.transaction.Status;
import javax.transaction.UserTransaction;

import com.zfabrik.util.threading.ThreadUtil;

/**
 * This class provides utilities around transaction management with Z2's built-in JTA implementation.
 * 
 * @author hb
 *
 */
public class TransactionUtil {

	/**
	 * Use this method to execute work in a transaction. This is just a utility that helps implementing
	 * for example Transaction Servlet Filters that begin and end transactions at the request boundaries.
	 * <p>
	 * If there is an ongoing transaction already, this method will not begin a new one. Any exception caught in that case will
	 * set that new transaction to rollback. 
	 * <p>
	 * See also {@link ThreadUtil#confine(Callable, Class)} on how to handle checked exception cases
	 * gracefully.
	 * 
	 * @param ut the {@link UserTransaction} instance of the transaction management system in place
	 * @param callable The {@link Callable} implementation that defines the work.
	 * @return The result of the callable invocation.
	 * @throws Exception
	 */
	public static <V> V run(UserTransaction ut, Callable<V> callable) throws Exception {
		if (ut.getStatus()!=Status.STATUS_ACTIVE) {
			ut.begin();
			try {
				return callable.call();
			} catch (Exception e) {
				logger.log(Level.WARNING,"Error at transaction boundary. Setting transaction to rollbackonly",e);
				ut.setRollbackOnly();
				throw e;
			} finally {
				if (ut.getStatus()==Status.STATUS_ACTIVE) {
					ut.commit();
				} else {
					ut.rollback();
				}
			}
		} else {
			// there is a tx ongoing already. Simply work.
			return callable.call();
		}
	}
	
	/**
	 * Use this method to execute work in a transaction. This is just a utility that helps implementing
	 * for example Transaction Servlet Filters that begin and end transactions at the request boundaries.
	 * <p>
	 * If there is an ongoing transaction already, this method will not begin a new one. Any exception caught in that case will
	 * set that new transaction to rollback.
	 * <p>
	 * Unlike {@link #run(UserTransaction, Callable)} this method does not declare a throws clause. Any runtime exception caught will be 
	 * rethrown, any other exception will be wrapped in a {@link RuntimeException}.
	 * 
	 * @param ut the {@link UserTransaction} instance of the transaction management system in place
	 * @param runnable The {@link Runnable} implementation that defines the work.
	 */
	public static void run(UserTransaction ut, Runnable runnable) {
		try {
			if (ut.getStatus()!=Status.STATUS_ACTIVE) {
				ut.begin();
				try {
					runnable.run();
				} catch (Exception e) {
					logger.log(Level.WARNING,"Error at transaction boundary. Setting transaction to rollbackonly",e);
					ut.setRollbackOnly();
					throw e;
				} finally {
					if (ut.getStatus()==Status.STATUS_ACTIVE) {
						ut.commit();
					} else {
						ut.rollback();
					}
				}
			} else {
				// there is a tx ongoing already. Simply work.
				runnable.run();
			}
		} catch (Exception e) {
			if (e instanceof RuntimeException) {
				throw (RuntimeException)e;
			}
			throw new RuntimeException(e);
		}
	}
	

	
	private final static Logger logger = Logger.getLogger(TransactionUtil.class.getName());
}
