/*
 * z2-Environment
 * 
 * Copyright(c) ZFabrik Software GmbH & Co. KG
 * 
 * contact@zfabrik.de
 * 
 * http://www.z2-environment.eu
 */
package com.zfabrik.impl.swing.components;

import java.awt.Font;
import java.lang.reflect.InvocationTargetException;
import java.util.logging.Logger;

import javax.swing.SwingUtilities;
import javax.swing.UIManager;

import com.zfabrik.components.IComponentDescriptor;
import com.zfabrik.components.java.JavaComponentUtil;
import com.zfabrik.resources.IResourceHandle;
import com.zfabrik.resources.ResourceBusyException;
import com.zfabrik.resources.provider.Resource;
import com.zfabrik.util.runtime.Foundation;
import com.zfabrik.work.WorkUnit;
import com.zfabrik.workers.home.IWorkersManager;

public class SwingApplicationResource extends Resource implements Runnable {

	private String name;
	// private IComponentDescriptor desc;

	private boolean loaded;
	private Resource worker;
	
	public SwingApplicationResource(String name) {
		this.name = name;
	}

	@Override
	public <T> T as(Class<T> clz) {
		this.load();
		if (Foundation.isWorker()) {
			if (Runnable.class.equals(clz)) {
				return clz.cast(this);
			}
		} else {
			return this.worker.as(clz);
		}
		return null;
	}
	
	public synchronized void invalidate() throws ResourceBusyException {
		try {
			if (loaded) {
				if (Foundation.isWorker()) {
					// then by dependency, the worker soul will go down
				} else {
					this.worker.invalidate();
				}
			}
		} finally {
			this.worker = null;
			loaded = false;
		}
	}

	public synchronized void load() {
		if (!loaded) {
			// IComponentDescriptor desc =
			// IComponentsManager.INSTANCE.getComponent(this.name);
			if (Foundation.isWorker()) {
				// on the worker side, we prepare the runnable that executes
				// the app and so there is nothing particular to do right here.
			} else {
				// on the home side, we create a worker resource and wrap it and
				// make us dependant
				if (this.worker == null) {
					this.worker = IWorkersManager.INSTANCE.createWorkerResource(handle(), this.name);
					this.worker.init(handle());
				}
				handle().adjust(0, Long.MAX_VALUE, IResourceHandle.HARD);
				// We want to be dependant on the java component (as that should kill the worker!)
				handle().addDependency(JavaComponentUtil.getJavaComponent(this.name));				
			}
			this.loaded = true;
		}
	}

	public void run() {
		// make the default fonts more agreeable.
		_choosePlainFont("Button.font");
		_choosePlainFont("Menu.font");
		_choosePlainFont("MenuItem.font");
		_choosePlainFont("TabbedPane.font");
		_choosePlainFont("List.font");
		_choosePlainFont("Label.font");
		_choosePlainFont("ComboBox.font");
		_choosePlainFont("Panel.font");
		_choosePlainFont("TitledBorder.font");
		_choosePlainFont("CheckBox.font");

		// a swing app must expose a Runnable. That is called in a Swing Thread.
		// It should not block.
		// the Swing app should set the default close operation to EXIT

		final Runnable ri = JavaComponentUtil.loadImplementationFromJavaComponent(
				this.name, 
				IComponentDescriptor.COMPONENT_CLZ, 
				handle(),
				Runnable.class
		);

		Runnable r = new Runnable() {
			public void run() {
				WorkUnit.work(ri);
			}
		};
		
		try {
			SwingUtilities.invokeAndWait(r);
		} catch (InterruptedException e) {
			throw new RuntimeException("Swing worker MAIN invocation dispatch failed", e);
		} catch (InvocationTargetException e) {
			throw new RuntimeException("Failed to initialize swing application", e.getCause());
		}
	}

	private static void _choosePlainFont(String key) {
		Font f = UIManager.getFont(key);
		if (f != null) {
			UIManager.put(key, new Font(f.getFamily(), Font.PLAIN, f.getSize()));
		} else {
			logger.warning("Failed to adjust font " + key);
		}
	}

	private final static Logger logger = Logger.getLogger(SwingApplicationResource.class.getName());
}
